import { world, ItemStack, EnchantmentType, BlockPermutation } from "@minecraft/server"

export const backpack = new class{
    //Spawn Backpack
    spawnBackpack(use, type){
        const block = use.source.getBlockFromViewDirection({maxDistance: 6, includeLiquidBlocks: true, includePassableBlocks: false})
        if(!block) return
        const pos = block.block.location
        const direction = {
            "North": [ {x: pos.x +0.5, y: pos.y, z: pos.z -0.5}, 0 ],
            "South": [ {x: pos.x +0.5, y: pos.y, z: pos.z +1.5}, 1 ],
            "East": [ {x: pos.x +1.5, y: pos.y, z: pos.z +0.5}, 2 ],
            "West": [ {x: pos.x -0.5, y: pos.y, z: pos.z +0.5}, 3 ],
            "Up": [ {x: pos.x +0.5, y: pos.y +1, z: pos.z +0.5}, Math.floor(Math.random() * 4) ],
            "Down": [ {x: pos.x +0.5, y: pos.y -1, z: pos.z +0.5}, Math.floor(Math.random() * 4) ]
        }
        const backpackBlock = world.getDimension(use.source.dimension.id).getBlock(direction[block.face][0])
        if(backpackBlock.typeId !== "minecraft:air" && backpackBlock.typeId !== "minecraft:water") return
        backpackBlock.setPermutation(BlockPermutation.resolve(`bp:backpack_${type}_block`, {"bp:dir": direction[block.face][1]}))
        const backpack = use.source.dimension.spawnEntity("bp:backpack", direction[block.face][0])
        backpack.nameTag = use.itemStack.nameTag ? use.itemStack.nameTag : "entity.bp:backpack.name"
        backpack.setDynamicProperty("pos", direction[block.face][0])
        use.itemStack.getDynamicProperty("backpackInventory") ? this.fillBackpack(use.itemStack.getDynamicProperty("backpackInventory"), backpack) : null
        use.source.runCommand(`clear @s bp:backpack_${type}_item 0 1`)
    }

    //Fill Backpack with Itens
    fillBackpack(inventoryInfo, backpack){
        const inventory = JSON.parse(inventoryInfo)
        const inv = backpack.getComponent("minecraft:inventory").container
        for(let i = 0; i < 27; i++){
            if(inventory[i]){
                const slot = inventory[i]
                const item = new ItemStack(slot.typeId, slot.amount)
                const durability = item.getComponent("minecraft:durability")
                const enchantments = item.getComponent("minecraft:enchantable")
                item.nameTag = slot.nameTag ? slot.nameTag : null
                durability ? durability.damage = durability.damage + slot.durability : null
                if(enchantments) if(slot.enchantments.length > 0) for(let i2 = 0; i2 < slot.enchantments.length; i2++){ enchantments.addEnchantment({"level": slot.enchantments[i2]["level"], "type": new EnchantmentType(slot.enchantments[i2]["type"]["id"])}) }
                if(slot.dynamicProperty) for(let i2 = 0; i2 < slot.dynamicProperty.length; i2++){ item.setDynamicProperty(slot.dynamicProperty[i2].id, slot.dynamicProperty[i2].property) }
                item.keepOnDeath = slot.keepOnDeath
                item.setCanDestroy(slot.canDestroy)
                item.setCanPlaceOn(slot.canPlaceOn)
                item.setLore(slot.lore)
                inv.setItem(i, item)
            }
        }
    }

    //Player Get Backpack
    "bp:backpack"(entity){
        let backpack = new ItemStack(`bp:backpack_${world.getDimension(entity.dimension.id).getBlock({...entity.location}).typeId.replace('bp:backpack_', '').replace('_block', '')}_item`)
        backpack.setDynamicProperty("backpackInventory", JSON.stringify(this.getItens(entity.getComponent("minecraft:inventory").container, entity)))
        backpack.nameTag = entity.nameTag !== "entity.bp:backpack.name" ? entity.nameTag : null
        world.getDimension(entity.dimension.id).getBlock({...entity.location}).setType("minecraft:air")
        world.getDimension(entity.dimension.id).spawnItem(backpack, {...entity.location})
        entity.triggerEvent("despawn")
    }

    //Get Itens on the Backpack
    invalidItens = ["bp:backpack_", "_shulker_box"]
    getItens(inventory, backpack){
        let itens = []
        for(let i = 0; i < 27; i++){
            if(!inventory.getItem(i)){ itens.push(undefined) } else {
                const item = inventory.getItem(i)
                if(!this.invalidItens.find(invalid => item.typeId.includes(invalid))){
                    const durability = item.getComponent("minecraft:durability")
                    const enchantments = item.getComponent("minecraft:enchantable")
                    itens.push({
                        typeId: item.typeId,
                        nameTag: item.nameTag,
                        amount: item.amount,
                        durability: durability ? durability.damage : undefined,
                        enchantments: enchantments ? enchantments.getEnchantments() : undefined,
                        dynamicProperty: item.getDynamicPropertyIds().length > 0 ? this.getDynamicProperty(item, item.getDynamicPropertyIds()) : undefined,
                        keepOnDeath: item.keepOnDeath,
                        canDestroy: item.getCanDestroy(),
                        canPlaceOn: item.getCanPlaceOn(),
                        lore: item.getLore()
                    })
                } else {
                    itens.push(undefined)
                    backpack.dimension.spawnItem(item.clone(), {...backpack.location})
                }
            }
        }
        return itens
    }

    //Get Dynamic Property on the Itens
    getDynamicProperty(item, properties){
        let dynamics = []
        for(let i = 0; i < properties.length; i++){ dynamics.push({id: properties[i], property: item.getDynamicProperty(properties[i])}) }
        return dynamics
    }

    "bp:paint"(ev){
        const block = world.getDimension(ev.sourceEntity.dimension.id).getBlock({...ev.sourceEntity.location})
        block.setPermutation(BlockPermutation.resolve(`bp:backpack_${ev.message}_block`, {"bp:dir": block.permutation.getState("bp:dir")}))
    }

    "bp:return"(ev){ ev.sourceEntity.teleport({...ev.sourceEntity.getDynamicProperty("pos")}) }
}